# coding: utf8

import collections.abc as c_abc
import logging
import pathlib
from typing import Dict, List, Union

from pydub import AudioSegment

from putao import internal, voicebank
from putao.exceptions import TrackError, ProjectError

_log = logging.getLogger("putao")


class Track:
    """A track (sequence of notes).

    Args:
        voicebank: The voicebank object to use to render notes.
    """

    def __init__(self, voicebank: voicebank.Voicebank):
        self._notes: List[internal.NoteBase] = []
        self.voicebank = voicebank

    def note(self, phoneme: str, pitch: int, duration: int):
        """Add a note.

        Args:
            phoneme: The syllable to sing.
            pitch: The absolute semitone value of the note.
            duration: How long to hold the note for (in miliseconds).
        """
        if phoneme not in self.voicebank:
            raise TrackError(f"'{phoneme}' does not exist in the voicebank")

        self._notes.append(internal.Note(self.voicebank[phoneme], pitch, duration))

    def rest(self, duration: int):
        """Add a rest (break in-between notes).

        Args:
            duration: How long to rest for.
        """
        self._notes.append(internal.Rest(duration))

    def render(self) -> AudioSegment:
        """Render all notes sequentially to an audio segment.

        Returns:
            The audio segment.
        """
        track_render = AudioSegment.empty()
        total = len(self._notes)

        for count, note in enumerate(self._notes, start=1):

            timestamp = len(track_render)
            preutter = 0
            overlap = 0

            try:
                next_note = self._notes[count + 1]
                if isinstance(next_note, internal.Note):
                    preutter = next_note.entry.preutterance
                    overlap = next_note.entry.overlap
            except IndexError:
                # no more notes
                pass

            _log.debug(
                "rendering note %s of %s (track duration: %ss)",
                count,
                total,
                timestamp / 1000,
            )

            render = note.render(preutter, overlap, self.voicebank.pitch)

            # extend final render (so overlay won't be truncated)
            track_render += AudioSegment.silent(len(render) - overlap)
            track_render = track_render.overlay(render, position=timestamp - overlap)

        return track_render

    def dump(self) -> List[dict]:
        notes = []

        for note in self._notes:
            dump = note.dump()
            if isinstance(note, internal.Note):
                dump.update(
                    {
                        "pitch": note.pitch,
                        "phoneme": note.entry.alias,
                    }
                )

            notes.append(dump)

        return notes

    def load(self, note_dumps: List[dict]):
        self._notes.clear()

        for dump in note_dumps:
            if dump["type"] == "note":
                self.note(*[dump[arg] for arg in ("phoneme", "pitch", "duration")])
            elif dump["type"] == "rest":
                self.rest(dump["duration"])


class Project(c_abc.MutableMapping):
    """A project (a.k.a song).

    First create a new track:

    proj = Project(".", 49)  # use the voicebank in curdir, pitched to C4
    track = proj.new_track("lead")

    If you want to access any existing tracks, use dict notation:

    existing_track = proj["lead"]

    You can also delete tracks:

    del proj["lead"]

    Args:
        vb_path: The path to the voicebank.
            The voicebank must be in UTAU format, and must have a oto.ini file.
        pitch: The pitch of the voicebank in semitones (as a int).
    """

    def __init__(self, vb_path: Union[str, pathlib.Path], pitch: int):
        self.voicebank = voicebank.Voicebank(vb_path, pitch)
        self.tracks: Dict[str, Track] = {}

    def __getitem__(self, name):
        return self.tracks[name]

    def __setitem__(self, name, value):
        raise ProjectError("can't directly set tracks: use .new_track() instead")

    def __delitem__(self, name):
        del self.tracks[name]

    def __iter__(self):
        return iter(self.tracks)

    def __len__(self):
        return len(self.tracks)

    def new_track(self, name: str) -> Track:
        """Create a new track.

        Args:
            name: The track name.

        Returns:
            The newly created track.

        Raises:
            ProjectError, if it already exists.
        """

        if name in self.tracks:
            raise ProjectError(f"track already exists: '{name}'")

        track = Track(self.voicebank)
        self.tracks[name] = track

        _log.debug("created new track %s", name)

        return track

    def render(self, path: Union[str, pathlib.Path]):
        """Render all tracks to a wavfile.

        Args:
            path: The file to render to.
        """

        project_render = AudioSegment.empty()
        for name, track in self.tracks.items():
            _log.debug("rendering track %s", name)
            render = track.render()

            project_len = len(project_render)
            render_len = len(render)

            if project_len < render_len:
                project_render += AudioSegment.silent(render_len - project_len)

            project_render = project_render.overlay(render)

        project_render.export(path, format="wav")

    def load(self, data: dict):
        """Load a project from a dict.

        Args:
            data: The project dict to load.
        """
        for name, notes in data["tracks"].items():
            track = self.new_track(name)
            track.load(notes)

    def dump(self) -> dict:
        """Dump a project to a dict.

        Returns:
            The project data as a dict.
        """
        data = {}
        for name, track in self.tracks.items():
            data[name] = track.dump()

        return data



# coding: utf8
"""Any classes/functions here should NOT be used outside of putao.
This contains the low-level parts of the rendering engine, and may change at any time.

You have been warned.
"""

import abc

from pydub import AudioSegment

from putao import utils, voicebank


class NoteBase(abc.ABC):
    def __init__(self, duration: int):
        self.duration = duration

    @abc.abstractmethod
    def render(self, preutter: int, overlap: int, pitch: int) -> AudioSegment:
        return

    def dump(self) -> dict:
        return {"type": self.__class__.__name__.lower(), "duration": self.duration}


class Rest(NoteBase):
    def render(self, preutter, overlap, pitch):
        return AudioSegment.silent(self.duration - preutter + overlap)


class Note(NoteBase):
    def __init__(self, entry: voicebank.Entry, pitch: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.entry = entry
        self.pitch = pitch

    def render(self, preutter, overlap, pitch):
        audio = AudioSegment.from_file(self.entry.wav)

        c_start = self.entry.offset
        c_end = c_start + self.entry.consonant
        consonant = audio[c_start:c_end]

        v_start = c_end
        v_end = len(audio) - self.entry.cutoff
        vowel = audio[v_start:v_end]

        phoneme_duration = len(consonant) + len(vowel)
        # preutterances extend this note into the previous one
        actual_duration = self.duration + self.entry.preutterance

        if actual_duration < phoneme_duration:
            # just cut off the end of the phoneme
            render = (consonant + vowel)[: self.duration]

        else:
            # calculate how much time will be used to loop the vowel.
            vowel_loop_dur = actual_duration - len(consonant)
            vowel_loop = AudioSegment.silent(vowel_loop_dur).overlay(vowel, loop=True)

            render = consonant + vowel_loop

        return utils.pitch_shift(
            render[: len(render) - preutter + overlap], self.pitch - pitch
        )



# coding: utf8

import collections
import hashlib
import math
import re

from pydub import AudioSegment

from putao.exceptions import ConversionError

RE_NOTE = re.compile(r"^((?i:[cdefgab]))([#b])?(\d+)$")

# in semitones
KEYS = {"c": 1, "d": 3, "e": 5, "f": 6, "g": 8, "a": 10, "b": 12}
NUM_KEYS = {v: k for k, v in KEYS.items()}

_note_length = collections.namedtuple(
    "note_length", "whole half quarter eighth sixteenth thirty_second sixty_fourth"
)
NOTE_LENGTH = _note_length(1, 2, 4, 8, 16, 32, 64)

SAMPLE_RATE = 44100
# stereo audio
CHANNELS = 2


class Pitch:
    """A numerical representation of a musical pitch.
    This class mainly serves to convert between different formats.

    Args:
        note: The pitch in scientific pitch notation, i.e C4 (key C, octave 4), C#4 (sharp), Cb4 (flat).
        semitone: The pitch in absolute number of semitones from C0, i.e 1 (C0), 49 (C4).
        hz: The pitch in Hertz, i.e 440 hz (C4).

    Raises:
        ValueError, if exactly one of 'note', 'semitone' or 'hz' is not specified.
        ConversionError, if the format/input is invalid.
    """

    def __init__(self, **kwargs):
        ((fmt, value),) = kwargs.items()

        if fmt == "semitone":
            self.semitone = value
        else:
            self.semitone = getattr(self, f"_from_{fmt}")(value)

        if self.semitone is None:
            raise ConversionError(f"invalid input: {value}")

    @property
    def note(self) -> str:
        key_number = self.semitone % 12
        try:
            key = str(NUM_KEYS[key_number])
        except KeyError:
            key = f"{NUM_KEYS[key_number - 1]}#"

        return f"{key.upper()}{self.semitone // 12}"

    @classmethod
    def _from_note(cls, note):
        try:
            key, accidental, octave = RE_NOTE.findall(note)[0]
        except IndexError:
            return None

        semitone = KEYS[key.lower()]
        if accidental == "#":
            semitone += 1
        elif accidental == "b":
            semitone -= 1

        return (int(octave) * 12) + semitone

    # A4 is note 58 here because the semitone is absolute (from C0, semitone = 1).
    @property
    def hz(self) -> float:
        return math.pow(math.pow(2, 1 / 12), self.semitone - 58) * 440

    @classmethod
    def _from_hz(cls, hz):
        return int(12 * math.log2(hz / 440) + 58)


def filesafe(name: str) -> str:
    """Hash name with the SHA1 algorithm, returning a slug suitable for a filename."""

    return hashlib.sha1(name.encode("utf8")).hexdigest()


def duration(length: int, bpm: int) -> int:
    """Calculate the duration of a note in miliseconds (how long it takes to play).
    One beat is defined as a quarter note.

    Args:
        length: A note length value from the NOTE_LENGTH namedtuple,
            i.e NOTE_LENGTH.whole (one note), NOTE_LENGTH.half (half note), etc.
        bpm: The tempo of the note in beats per minute.
    """
    if length not in NOTE_LENGTH:
        return 0

    return int(((60 / bpm) * (NOTE_LENGTH.quarter / length)) * 1000)


def pitch_shift(audio: AudioSegment, semitones: int) -> AudioSegment:
    """Shift the pitch of the audio segment by semitones.
    (https://github.com/jiaaro/pydub/issues/157#issuecomment-252366466)

    Args:
        audio: The pydub audio segment to pitch-shift.
        semitones: How many semitones to shift (negative values will shift down).

    Returns:
        The pitch-shifted segment.
    """

    octaves = semitones / 12
    new_sr = int(audio.frame_rate * (2 ** octaves))

    pitch_shifted = audio._spawn(audio.raw_data, overrides={"frame_rate": new_sr})
    return pitch_shifted.set_frame_rate(SAMPLE_RATE)



# coding: utf8

import io
from dataclasses import dataclass
from typing import List

import mido


@dataclass
class Note:
    # absolute start/end
    start: int
    end: int
    pitch: int
    tpb: int
    tempo: int

    def dump(self):
        return {
            "type": "note",
            "pitch": self.pitch,
            "duration": mido.tick2second(self.end - self.start, self.tpb, self.tempo),
        }


def _collides(note1: Note, note2: Note) -> bool:
    return (note2.start <= note1.start < note2.end) or (
        note2.start < note1.end <= note2.end
    )


def _tempo(meta):
    clock = 0
    changes = []
    for msg in meta:
        clock += msg.time
        if msg.type == "set_tempo":
            changes.append((clock, msg.tempo))

    return changes


class Song:
    def __init__(self, data: bytes):
        buf = io.BytesIO(data)
        mid = mido.MidiFile(file=buf)

        self.notes: List[Note] = []

        meta = mid.tracks[0]
        if any(msg.type == "note_on" for msg in meta):
            # lead is also meta
            lead = meta
        else:
            lead = mid.tracks[5]

        tempos = _tempo(meta)

        clock = 0
        next_clock = 0
        current_tempo = tempos[0][1]
        tempo_counter = 0

        for count, msg in enumerate(lead):
            clock += msg.time

            if msg.type != "note_on":
                continue

            next_clock = clock
            for next_count, next_msg in enumerate(lead[count + 1 :]):
                next_clock += next_msg.time
                if next_msg.type == "note_off" and next_msg.note == msg.note:
                    break

            # apply change in tempo (if any)
            if not (tempo_counter >= len(tempos)):
                tempo_clock, tempo = tempos[tempo_counter]
                if clock <= tempo_clock <= next_clock:
                    current_tempo = tempo
                    tempo_counter += 1

            self.notes.append(
                Note(clock, next_clock, msg.note, mid.ticks_per_beat, current_tempo)
            )

    def dump(self) -> List[dict]:
        track: List[dict] = []

        for count, note in enumerate(self.notes):
            prev_note = self.notes[count - 1]
            note_dump = note.dump()

            if note.start > 0:

                # there is a rest between this note and the previous one
                if count >= 1:
                    interval = note.start - prev_note.end
                else:
                    # first note, but there is a pause at the start
                    interval = note.start

                rest = mido.tick2second(interval, note.tpb, note.tempo)

                if rest < 0:
                    # overlapping notes
                    note_dump["overlap"] = abs(rest)

                else:
                    track.append(
                        {
                            "type": "rest",
                            "rest": rest,
                        }
                    )

            track.append(note_dump)

        return track


def loads(data: bytes):
    return Song(data).dump()



# coding: utf8
"""mmlx (Music Macro Language eXtended) parser and interpreter.

TODO
- add support for ADSR envelopes through 'v' command.

(Anything enclosed in '[]' is optional.)

Core syntax:

    lengths below may be one of '1', '2', '4', '8', '16', '32', '64'
    (1 -> whole note, 2 -> half note, 4 -> quarter note, 8 -> eighth note, etc.)

    (key)[accidental][length]
        Play a note.

        key must be one of the letters 'abcdefg',
        and accidental may be '+', '#' (sharp) or '-' (flat).
        i.e 'c2'

    p[length] / r[length]
        Pause/rest playing notes.

    o(octave)
        Change the octave of the notes.

        octave must be a integer.

    > / <
        Shift octave up or down, respectively.

    l(length)
        Change the length of the notes.

    t(tempo)
        Change the tempo of the notes.

        tempo must be an integer, in beats per minute.

Extended syntax:

    #[comment]
        A comment.
        Comments can appear anywhere in a line (i.e after commands).
        All text between the '#' until the next newline is ignored.

    @(extension) [args] {...}
        General syntax for an extension.
        Extensions can modify/apply changes to the notes within its scope (the '{...}').

    @track (trackname) {...}
        Add scoped notes to trackname.
        Any notes outside of a @track scope are implictly added to the 'global' track.

    @loop (times) {...}
        Repeat scoped notes by number of times.
        times must be an integer.

    @lyrics (phonemes) {...}
        Add lyrics (split by spaces) to the scoped notes.
        If there are not enough phonemes, the last one will be repeated for the rest of the notes.
"""

import collections
from dataclasses import dataclass
from typing import Any, Dict, List

import pyparsing as pp

from putao import utils

pp_u = pp.pyparsing_unicode


PROPMAP = {"o": "octave", "l": "length", "t": "tempo"}


@dataclass
class Token:
    name: str
    value: Any
    loc: int


def _note(loc, tk):
    try:
        key, length = tk[0]
        length = int(length)
    except ValueError:
        key = tk[0][0]
        length = 0

    # the none is a placeholder for a phoneme.
    return Token("note", (key, length, None), loc)


def _rest(loc, tk):
    return Token("rest", tk[0], loc)


def _prop(loc, tk):
    try:
        prop, value = tk[0]
        value = int(value)
        shift = False

    except ValueError:
        prop = "o"
        value = tk[0]

        if value == "<":
            value = -1
        elif value == ">":
            value = 1

        shift = True

    return Token("prop", (prop, value, shift), loc)


def _oct_shift(loc, tk):
    return Token("oct_shift", 1 if tk[0] == ">" else -1, loc)


def _scope_header(loc, tk):
    return Token("scope", list(tk[0]), loc)


# core MML syntax
def _mml_syntax():
    key = pp.Combine(pp.Char("cdefgab") + pp.Optional(pp.Char("+-#")))
    key.setParseAction(lambda tk: tk[0].replace("+", "#").replace("-", "b"))

    length = pp.Optional(pp.Word(pp.nums))

    note = pp.Group(key + length)
    note.setParseAction(_note)

    rest = pp.Combine((pp.Suppress("r") | pp.Suppress("p")) + length)
    rest.setParseAction(_rest)

    prop_num = pp.Word(pp.nums)

    prop = pp.Group(pp.Char("olt") + prop_num)
    prop.setParseAction(_prop)

    oct_shift = pp.Char("<>")
    oct_shift.setParseAction(_oct_shift)

    mml = (note | rest | prop | oct_shift)[1, ...]

    # eXtended syntax
    comment = pp.Literal("#") + pp.restOfLine

    scope = pp.Forward()
    scope_header = pp.Group(
        pp.Combine(pp.Suppress("@") + pp.Word(pp.alphas))
        + (pp.Word(pp.alphanums) | pp.Word(pp_u.Japanese.printables))[...]
    )
    scope_header.setParseAction(_scope_header)

    inner_scope = pp.Group(pp.Suppress("{") + scope + pp.Suppress("}"))

    scope << ((scope_header + inner_scope) | scope_header | mml)[1, ...]

    mmlx = (mml | comment | scope)[1, ...]
    mmlx.ignore(comment)

    return mmlx


Parser = _mml_syntax()


class Interpreter:
    def __init__(self, mml: str):
        self._tokens = Parser.parseString(mml)
        # per-track variables
        self._props = {
            "global": {
                "octave": 4,
                "length": utils.NOTE_LENGTH.quarter,
                "tempo": 120,
            }
        }

        self.tracks: Dict[str, List[dict]] = collections.defaultdict(list)
        self.current_track = "global"

    def _prop(self) -> dict:
        return self._props.setdefault(self.current_track, self._props["global"])

    def note(self, token, track):
        key, length, phoneme = token.value

        note = {
            "type": "note",
            # In mml, we use a 'global' octave so we have to calculate the semitone here.
            "pitch": utils.Pitch(note=f"{key}{track['octave']}").semitone,
            "duration": utils.duration(length or track["length"], track["tempo"]),
        }

        if phoneme is not None:
            note["phoneme"] = phoneme

        self.tracks[self.current_track].append(note)

    def rest(self, token, track):
        length = token.value

        self.tracks[self.current_track].append(
            {
                "type": "rest",
                "duration": utils.duration(length or track["length"], track["tempo"]),
            }
        )

    def prop(self, token, track):
        prop, value, shift = token.value
        if shift:
            track[PROPMAP[prop]] += value
        track[PROPMAP[prop]] = int(value)

    def oct_shift(self, token, track):
        track["octave"] += token.value

    def scope_track(self, args, tokens, track):
        self.current_track = args[0]

    def scope_loop(self, args, tokens, track):
        times = int(args[0]) - 1
        for _ in range(times):
            tokens.extend(tokens)

    def scope_lyrics(self, args, tokens, track):
        lyrics = iter(args)
        last_phoneme = args[-1]

        for token in tokens:
            if token.name == "scope":
                raise RuntimeError("scopes are not allowed within lyrics")

            if token.name == "note":
                key, length, _ = token.value
                token.value = (key, length, next(lyrics, last_phoneme))

    def _execute(self, tokens):
        counter = 0
        limit = len(tokens)

        while counter < limit:
            token = tokens[counter]
            track = self._prop()

            if token.name != "scope":
                getattr(self, token.name)(token, track)

            else:
                # the next token is actually a list of tokens under this scope
                # so take the next one
                try:
                    scope_tokens = tokens[counter + 1]
                except IndexError:
                    scope_tokens = []
                name, *args = token.value

                try:
                    getattr(self, f"scope_{name}")(args, scope_tokens, track)
                except AttributeError:
                    raise RuntimeError(f"{name}: no such extension exists")

                self._execute(scope_tokens)

                # skip the scoped tokens
                counter += 1

            counter += 1

    def execute(self):
        self._execute(self._tokens)
        return {"tracks": dict(self.tracks)}


def loads(data):
    itpr = Interpreter(data.decode("utf8"))
    return itpr.execute()



# coding: utf8

from dataclasses import dataclass
from typing import Literal

import mido


@dataclass
class Event:
    type: Literal["note", "rest", "set_tempo"]
    # start and end is absolute (midi ticks)
    start: int
    end: int
    pitch: int
    tpb: int
    tempo: int

    def dump(self):
        return {
            "type": self.type,
            "pitch": self.pitch,
            "duration": mido.tick2second(self.end - self.start, self.tpb, self.tempo),
        }

    def __eq__(self, event):
        if isinstance(event, Event):
            return self.pitch == event.pitch
        else:
            return NotImplemented



# coding: utf8
"""葡萄 (putao, grape): Poor man's UTAU."""

__version__ = "0.1.0a0"

import logging

from putao.core import Project  # noqa: W0611

logging.getLogger("sox").setLevel(logging.ERROR)
logging.basicConfig(level=logging.DEBUG)



# coding: utf8

from putao import core, backend, utils

vb_path = "./voicebank"

vb_pitch = utils.Pitch(
    note=input("what is the pitch of this voicebank? (i.e C4, D#5, etc.) > ")
).semitone

proj = core.Project(vb_path, vb_pitch)

file = input("path to the mml file you want to render > ")

with open(f"{file}.mml", "rb") as f:
    proj_data = backend.load(f, "mmlx")
    proj.load(proj_data)

proj.render(f"{file}.wav")



# coding: utf8
"""Backends provide an abstraction between a source of music notes and putao's song architecture.

Each backend must have a loads() function that accepts a single argument 'data' with type 'bytes'.
This function must return a dict of lists:

{  # tracks
    "track_name": [  # notes
        {
            'type': ...  # either 'note' or 'rest'
            'duration': ...  # how long to hold the note for
            'pitch': ...  # the absolute semitone value of the note, only needed if type is 'note'.
            'overlap': ...  # how much the note overlaps with the previous one. Optional (default:0.0)
        },
        ...
    ]
}
"""

import importlib
import pathlib
import sys
from typing import IO

BACKENDS = [
    p.stem for p in pathlib.Path(__file__).parent.glob("*.py") if p.stem != "__init__"
]


def loads(data: bytes, fmt: str) -> dict:
    """Load data according to fmt.

    Args:
        data: The music data to load.
        fmt: Which backend to use to load the music data.
            Available backends are in BACKENDS.

    Returns:
        The data as a project dict.
    """

    module = f"putao.backend.{fmt}"
    if module in sys.modules:
        backend = sys.modules[module]
    else:
        backend = importlib.import_module(module)

    return backend.loads(data)  # type: ignore


def load(fp: IO, *args) -> dict:
    """Load data from fp according to fmt.

    Args:
        fp: The file to read data from.
            It must be opened in bytes mode ('rb').
        *args: Passed to loads().

    Returns:
        The data as a project dict.
    """

    return loads(fp.read(), *args)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['putao', 'putao.backend']

package_data = \
{'': ['*'], 'putao': ['.ropeproject/*'], 'putao.backend': ['.ropeproject/*']}

install_requires = \
['chardet>=4.0.0', 'mido>=1.2.9', 'pydub>=0.25.1', 'pyparsing>=2.4.7']

setup(name='putao',
      version='0.1.0a0',
      description="葡萄 (putao, grape): Poor man's UTAU.",
      author='Ong Yong Xin',
      author_email='ongyongxin2020+github@gmail.com',
      url='https://github.com/ongyx/putao',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.6',
     )



# coding: utf8
"""UTAU voicebank interface.
NOTE: all time values are in miliseconds.
"""

from __future__ import annotations

import collections.abc as c_abc
import logging
import pathlib
import re
import zipfile
from dataclasses import dataclass
from typing import Union

import chardet

_log = logging.getLogger("putao")

RE_SYLLABLE = re.compile(r"(\w+\.wav)=(.+)" + (r",(-?\d+)" * 5))


@dataclass
class Entry:
    """An entry in the oto.ini config of an UTAU voicebank.
    (All time values are in miliseconds.)

    Args:
        wav: The absolute path to the wavfile.
        alias: The name of the phoneme to associate with the wavfile.
        offset: Unused length of time at the _start_ of the wavfile.
        consonant: Length of time of the consonant.
        cutoff: Unused length of time at the _end_ of the wavfile.
        preutterance: Length of time to extend the start of the note into the previous one.
            The previous note is shortened.
            This should be shorter than the overlap (below).
        overlap: Length of time to extend the previous note into the current note.
            This extension overlaps into the start of the current note.

    Attributes:
        See args.
    """

    wav: Union[str, pathlib.Path]
    alias: str
    offset: int
    consonant: int
    cutoff: int
    preutterance: int
    overlap: int

    @staticmethod
    def parse(entry: str) -> Entry:
        """Parse a line in an oto.ini file.

        Args:
            entry: The line to parse.

        Returns:
            The entry object.
        """
        wav, alias, *times = RE_SYLLABLE.findall(entry)[0]
        times = [int(t) for t in times]

        return Entry(wav, alias, *times)


class Voicebank(c_abc.Mapping):
    """An UTAU voicebank.

    Args:
        path: The path to the voicebank.
        pitch: The semitone value of the wavfiles in the voicebank.

    Attributes:
        path: See args.
        pitch: See args.
    """

    def __init__(self, path: Union[str, pathlib.Path], pitch: int):
        self.path = pathlib.Path(path)
        self.pitch = pitch

        self._wav_map = {}

        _log.debug("parsing oto.ini")

        # assuming the voicebank is already utf8...
        with (self.path / "oto.ini").open() as f:
            for _entry in f.readlines():
                entry = Entry.parse(_entry)

                # absolute path
                entry.wav = self.path / entry.wav
                self._wav_map[entry.alias] = entry

        _log.debug("parsed oto.ini")

    def __getitem__(self, key):
        return self._wav_map[key]

    def __iter__(self):
        return iter(self._wav_map)

    def __len__(self):
        return len(self._wav_map)


def unmojibake(text: Union[str, bytes]) -> str:
    """Re-encode text/bytes to UTF8 from Shift-JIS or other encodings.

    Args:
        text: The text to re-encode.

    Returns:
        The UTF8-ified text.
    """

    if isinstance(text, str):
        raw = text.encode("cp437")
    else:
        # already bytes
        raw = text

    try:
        return raw.decode("sjis")
    except UnicodeDecodeError:
        # use chardet to figure out encoding
        encoding = chardet.detect(raw)["encoding"]
        return raw.decode(encoding)


def is_utf8(zinfo: zipfile.ZipInfo) -> bool:
    """Check whether the filename of the zipinfo is utf8."""
    return bool(zinfo.flag_bits & 0x800)


def extract_voicebank(
    path: Union[str, pathlib.Path],
    to: Union[str, pathlib.Path],
    convert_newlines: bool = True,
):
    """Extract an UTAU voicebank zipfile, while decoding file(names) correctly.

    Args:
        path: The path to the zipfile.
        to: The folder to extract the zipfile to.
        convert_newlines: Whether or not to replace Windows-style newlines (CRLF) with *nix-style newlines (LF).
            Defaults to True.
    """
    path = pathlib.Path(path)
    to = pathlib.Path(to)

    with zipfile.ZipFile(path) as zf:
        for zinfo in zf.infolist():
            if not is_utf8(zinfo):
                # most voicebanks are either romanji (ASCII) or kanji/hirigana (SHIFT-JIS).
                zinfo.filename = unmojibake(zinfo.filename)

            full_path = to / zinfo.filename
            full_path.parent.mkdir(exist_ok=True)

            # decode any text files and extract manually
            if full_path.suffix[1:] in ("txt", "ini"):
                text = unmojibake(zf.read(zinfo))

                if convert_newlines:
                    text = text.replace("\r\n", "\n")

                full_path.write_text(text)

            else:
                zf.extract(zinfo, path=to)
